#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include "bufferedio.h"
#include "safeio.h"

int buffered_reader_init(BufferedReader * const br, int fd)
{
   if (br == NULL || fd < 0) {
      return -1;
   }
   if ((br->buf = malloc(STREAM_BUFFER_READ_SIZE)) == NULL) {
      return -1;
   }
   br->fd = fd;
   br->buf_pnt = br->buf_end = br->buf;
   
   return 0;
}

int buffered_reader_close(BufferedReader * const br, int really)
{
   if (br == NULL || br->fd < 0) {
      return -1;
   }
   if (really != 0) {
      close(br->fd);
   }
   if (br->buf != NULL) {
      free(br->buf);
      br->buf = NULL;
   }
   br->buf_pnt = br->buf_end = NULL;
   br->fd = -1;
   
   return 0;
}

ssize_t buffered_reader_read(BufferedReader * const br,
			     void *where_, size_t size)
{
   unsigned char *where = (unsigned char *) where_;
   ssize_t readen;
   
   if (size == (size_t) 0U || br->buf_pnt == NULL) {
      return (ssize_t) 0;
   }
   for(;;) {
      while (br->buf_pnt != br->buf_end) {
	 *where++ = *(br->buf_pnt++);
	 if (--size == (size_t) 0U) {
	    return (ssize_t) (where - (unsigned char *) where_);
	 }
      }
      br->buf_pnt = br->buf_end = NULL;
      if ((readen = safe_read(br->fd, br->buf, STREAM_BUFFER_READ_SIZE))
	  < (ssize_t) 0) {
	 return (ssize_t) -1;
      }
      if (readen == (ssize_t) 0) {
	 br->buf_pnt = NULL;
	 return (ssize_t) (where - (unsigned char *) where_);      
      }
      br->buf_pnt = br->buf;
      br->buf_end = br->buf + readen;
   }
}

int buffered_reader_get_line(BufferedReader * const br,
			     char *buf, size_t size)
{
   char c;
   ssize_t res;

   if (size < (size_t) 2U) {
      return -2;
   }
   for (;;) {
      size--;
      if (size == (size_t) 0U) {
	 *buf = 0;
	 return -3;
      }
      if ((res = buffered_reader_readchar(br, &c)) < (ssize_t) 0) {
	 *buf = 0;
	 return -1;
      } else if (res == 0) {
	 *buf = 0;
	 return 0;
      }
      *buf = c;
      buf++;
      if (c == '\n') {
	 break;
      }
   }
   *buf = 0;
   
   return 1;
}

int buffered_writer_init(BufferedWriter * const bw, int fd)
{
   if (bw == NULL || fd < 0) {
      return -1;
   }
   if ((bw->buf = malloc(STREAM_BUFFER_WRITE_SIZE)) == NULL) {
      return -1;
   }
   bw->fd = fd;
   bw->buf_pnt = bw->buf;
   bw->space_left = STREAM_BUFFER_WRITE_SIZE;
   
   return 0;
}

int buffered_writer_write(BufferedWriter * const bw,
			  const void * const where_, size_t size)
{
   const unsigned char *where = (const unsigned char *) where_;

   anew:
   if (size <= bw->space_left) {
      memcpy(bw->buf_pnt, where, size);
      bw->space_left -= size;
      bw->buf_pnt += size;
      return 0;
   }   
   if (bw->space_left > 0U) {
      memcpy(bw->buf_pnt, where, bw->space_left);
   }   
   if (safe_write(bw->fd, bw->buf, STREAM_BUFFER_WRITE_SIZE) < 0) {
      return -1;
   }
   size -= bw->space_left;
   where += bw->space_left;
   bw->buf_pnt = bw->buf;
   bw->space_left = STREAM_BUFFER_WRITE_SIZE;
   goto anew;   
}

int buffered_writer_flush(BufferedWriter * const bw)
{
   const size_t size = (size_t) (bw->buf_pnt - bw->buf);
   
   if (size != (size_t) 0U) {
      if (safe_write(bw->fd, bw->buf, size) < 0) {
	 return -1;
      }
      bw->buf_pnt = bw->buf;
      bw->space_left = STREAM_BUFFER_WRITE_SIZE;
   }   
   return 0;
}

int buffered_writer_close(BufferedWriter * const bw, int really)
{
   if (bw == NULL || bw->fd < 0) {
      return -1;
   }
   buffered_writer_flush(bw);
   if (really != 0) {
      close(bw->fd);   
   }
   if (bw->buf != NULL) {
      free(bw->buf);
      bw->buf = NULL;
   }   
   bw->buf_pnt = NULL;
   bw->space_left = (size_t) 0U;
   bw->fd = -1;
   
   return 0;
}
