#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include "lmail.h"
#include "cdb/cdb.h"
#include "lmail-queue.h"
#include "safeio.h"
#include "strtools.h"
#include "list.h"
#include "hashedlist.h"
#include "configfiles.h"
#include "bufferedio.h"

int configfile_read_hashed_list(const char * const file_name,
				HashedList * const hl)
{
   char line[LINE_MAX];
   char *line_trim = NULL;
   BufferedReader br;
   int fd;
   
   br.fd = -1;
   if ((fd  = open(file_name, O_RDONLY | O_NDELAY)) < 0) {
      D(perror("open"));
      goto exit_failure;
   }
   if (buffered_reader_init(&br, fd) < 0) {
      D(perror("buffered_reader_init"));
      close(fd);
      goto exit_failure;
   }
   D(printf("Reading control file [%s]\n" __ file_name));
   while (buffered_reader_get_line(&br, line, sizeof line) > 0) {
      chomp(line);
      line_trim = skip_spaces(line);
      trim_rspaces(line_trim);
      if (line_trim[0] != CONFIGFILES_COMMENT_CHAR && line_trim[0] != 0) {
	 if ((line_trim = strdup(line_trim)) == NULL) {
	    D(perror("malloc"));
	    continue;
	 }
	 lowerize(line_trim);
	 if (hashed_list_push(hl, line_trim, TRIVIAL_HASH(hl, line_trim)) < 0) {
	    D(perror("hashed_list_push"));
	 }
      }
   }
   buffered_reader_close(&br, 1);
   return 0;
   
   exit_failure:
   if (br.fd >= 0) {
      buffered_reader_close(&br, 1);
   }
   return -1;
}

int configfile_read_rcpthosts_list(const char * const file_name,
				   HashedList * const hl_strict,
				   HashedList * const hl_joker)
{
   char line[LINE_MAX];
   char *line_trim = NULL;
   BufferedReader br;
   int fd;
   
   br.fd = -1;
   if ((fd  = open(file_name, O_RDONLY | O_NDELAY)) < 0) {
      D(perror("open"));
      goto exit_failure;
   }
   if (buffered_reader_init(&br, fd) < 0) {
      D(perror("buffered_reader_init"));
      close(fd);
      goto exit_failure;
   }
   D(printf("Reading control file [%s]\n" __ file_name));
   while (buffered_reader_get_line(&br, line, sizeof line) > 0) {
      chomp(line);
      line_trim = skip_spaces(line);
      trim_rspaces(line_trim);
      if (line_trim[0] != CONFIGFILES_COMMENT_CHAR && line_trim[0] != 0) {
	 if (line_trim[0] == '.') {
	    if ((line_trim = strdup(line_trim)) == NULL) {
	       D(perror("malloc"));
	       continue;
	    }
	    lowerize(line_trim);
	    if (hashed_list_push(hl_joker, line_trim,
				 REVERSE_TRIVIAL_HASH(hl_joker, line_trim)) < 0) {
	       D(perror("hashed_list_push"));
	    }
	 } else {
	    if ((line_trim = strdup(line_trim)) == NULL) {
	       D(perror("malloc"));
	       continue;
	    }
	    lowerize(line_trim);
	    if (hashed_list_push(hl_strict, line_trim,
				 TRIVIAL_HASH(hl_strict, line_trim)) < 0) {
	       D(perror("hashed_list_push"));
	    }
	 }
      }
   }
   buffered_reader_close(&br, 1);
   return 0;
   
   exit_failure:
   if (br.fd >= 0) {
      buffered_reader_close(&br, 1);
   }
   return -1;
}

static char *configfile_read_line(const char * const file_name,
				  char *line, const size_t line_size)
{
   char *line_trim;
   BufferedReader br;
   int fd;

   br.fd = -1;
   if ((fd = open(file_name, O_RDONLY | O_NDELAY)) < 0) {
      D(perror("open"));
      goto exit_failure;
   }
   if (buffered_reader_init(&br, fd) < 0) {
      D(perror("buffered_reader_init"));
      close(fd);
      goto exit_failure;
   }
   D(printf("Reading control file [%s]\n" __ file_name));
   while (buffered_reader_get_line(&br, line, line_size) > 0) {
      chomp(line);
      line_trim = skip_spaces(line);
      trim_rspaces(line_trim);
      if (line_trim[0] != CONFIGFILES_COMMENT_CHAR && line_trim[0] != 0) {
	 buffered_reader_close(&br, 1);
	 return line_trim;
      }
   }
   exit_failure:
   if (br.fd >= 0) {
      buffered_reader_close(&br, 1);
   }
   return NULL;   
}

const char *configfile_read_string(const char * const file_name)
{   
   char line[LINE_MAX];
   char *string;
   
   if ((string = configfile_read_line(file_name, line, sizeof line)) == NULL) {
      return NULL;
   }
   if ((string = strdup(string)) == NULL) {
      D(perror("strdup"));
      return NULL;
   }
   D(printf("read_string from config file [%s] : [%s]\n" __
	    file_name __ string));
   
   return string;
}

unsigned long configfile_read_unsigned_long(const char * const file_name)
{
   char line[LINE_MAX];
   char *string;
   char *resultf;
   unsigned long result;
   
   if ((string = configfile_read_line(file_name, line, sizeof line)) == NULL) {
      return 0UL;
   }
   if ((string = strdup(string)) == NULL) {
      D(perror("strdup"));
      return 0UL;
   }
   result = strtoul(string, &resultf, 0);
   if (*string != 0 && *resultf == 0) {
      D(printf("read_unsigned_long from config file [%s] : [%lu]\n" __
	       file_name __ result));
      free(string);      
      return result;
   }
   free(string);   
   D(puts("No numeric input"));
   
   return 0UL;
}

int configfile_init_cdb(const char * const file_name, CDB * const cdb)
{
   int fd;
   
   if ((fd  = open(file_name, O_RDONLY | O_NDELAY)) < 0) {
      D(perror("open"));
      return -1;
   }
   D(printf("Reading control CDB database [%s]\n" __ file_name));
   cdb->fd = -1;
   cdb_init(cdb, fd);
   
   return 0;
}

int configfile_free_cdb(CDB * const cdb)
{
   if (cdb == NULL || cdb->fd < 0) {
      D(puts("configfile_free_cdb called with wrong args"));
      return -1;
   }
   cdb_free(cdb);   
   if (close(cdb->fd) < 0) {
      D(perror("close"));
      return -1;
   }
   cdb->fd = -1;
   
   return 0;
}

