#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "st/st.h"
#include "lmail.h"
#include "varstring.h"
#include "list.h"
#include "hashedlist.h"
#include "strtools.h"
#include "bufferedio.h"
#include "cdb/cdb.h"
#include "lmail-queue.h"
#include "lmail-smtpd.h"
#include "hacks.h"
#include "rfc.h"
#include "trigger.h"
#include "sequence.h"
#include "sync-queue.h"
#include "safeio.h"
#include "configfiles.h"
#include "checkaddr.h"

static const char *HEADER_RECEIVED = "received";
static const char *HEADER_DELIVERED = "delivered";
static char server_shutdown;
#ifdef LAZY_METADATA_SYNC
static int sync_semid = -1;
static SyncQueue global_sync_queue;
#endif
static ConfigFiles config;
int total_clients;

static void ipc_delete(void)
{
   D(puts("IPC removal"));
}

#ifdef __GNUC__
RETSIGTYPE sig(int sg) __attribute__ ((__noreturn__));
#endif

RETSIGTYPE sig(int sg)
{
   D(printf("Signal %d received\n" __ sg));
   signal(sg, SIG_DFL);
   ipc_delete();
   _exit(EXIT_FAILURE);
#if RETSIGTYPE != void
   return 0;
#endif
}

#ifdef __GNUC__
RETSIGTYPE sig_term(int sg) __attribute__ ((__noreturn__));
#endif

RETSIGTYPE sig_term(int sg)
{
   unsigned int grace = MAX_SHUTDOWN_TIME;
   
   D(printf("Signal %d received\n" __ sg));
   signal(sg, SIG_DFL);
   D(puts("Starting graceful abort"));
   server_shutdown = 1;
   while (grace != 0U) {
      sleep(1U);
      if (total_clients <= 0) {
	 break;
      }
      grace--;
   }
   D(puts("Graceful abort"));
   server_shutdown = 2;
   ipc_delete();
   _exit(EXIT_SUCCESS);
#if RETSIGTYPE != void
   return 0;
#endif
}

static int read_config_files(void)
{
   if ((config.smtpgreeting =
       configfile_read_string(CONTROL_PATH_SMTPGREETING)) == NULL) {
      config.smtpgreeting = PACKAGE "-" VERSION;
   }
   if ((config.databytes = (size_t) 
	configfile_read_unsigned_long(CONTROL_PATH_DATABYTES)) == (size_t) 0U) {
      config.databytes = (size_t) (UINT_MAX - 256U);
   }
   if ((config.smtpdport = (short)
	configfile_read_unsigned_long(CONTROL_PATH_SMTPDPORT)) == (short) 0) {
      config.smtpdport = (short) SMTP_PORT;
   }
   if ((config.max_hops = (unsigned int)
	configfile_read_unsigned_long(CONTROL_PATH_MAXHOPS)) == 0U) {
      config.max_hops = MAX_HOPS;
   }
   if ((config.max_clients = (int)
	configfile_read_unsigned_long(CONTROL_PATH_MAXCLIENTS)) == 0U) {
      config.max_clients = MAX_CLIENTS;
   }
   config.has_badmailfrom = 0;
   if (hashed_list_init(&config.badmailfrom, BADMAILFROM_SPLIT) < 0) {
      perror("Failure with " CONTROL_PATH_BADMAILFROM);
      return -1;      
   } else {
      if (configfile_read_hashed_list(CONTROL_PATH_BADMAILFROM, 
				      &config.badmailfrom) == 0) {
	 config.has_badmailfrom = 1;
      }
   }
   config.open_relay = 0;
   if (hashed_list_init(&config.rcpthosts_strict, RCPTHOSTS_SPLIT) < 0 ||
       hashed_list_init(&config.rcpthosts_joker, RCPTHOSTS_SPLIT) < 0) {
      perror("Failure with " CONTROL_PATH_RCPTHOSTS);
      return -1;
   } else {
      if (configfile_read_rcpthosts_list(CONTROL_PATH_RCPTHOSTS, 
					 &config.rcpthosts_strict,
					 &config.rcpthosts_joker) != 0) {
	 config.open_relay = 1;
	 hashed_list_destroy(&config.rcpthosts_strict);
	 hashed_list_destroy(&config.rcpthosts_joker);	 
      }
   }
   config.morercpthosts.fd = -1;
   configfile_init_cdb(CONTROL_PATH_MORERCPTHOSTS, &config.morercpthosts);
   
   config.has_relayclients = 0;
   if (hashed_list_init(&config.relayclients, RELAYCLIENTS_SPLIT) < 0) {
      perror("Failure with " CONTROL_PATH_RELAYCLIENTS);
      return -1;      
   } else {
      if (configfile_read_hashed_list(CONTROL_PATH_RELAYCLIENTS, 
				      &config.relayclients) == 0) {
	 config.has_relayclients = 1;
      }
   }   
   
   return 0;
}

static inline int say(State * const state, const char *x)
{
   return buffered_writer_write(state->client_bw, x, strlen(x));
}

static inline int say_flush(State * const state, const char *x)
{
   int ret;
   
   if ((ret = say(state, x)) < 0) {
      return ret;
   }
   return buffered_writer_flush(state->client_bw);
}

static int cmd_quit(char *arg, State * const state)
{
   (void) arg;
   (void) state;  
   say_flush(state, _("221 cu") CRLF);
   
   return CMD_RETURN_QUIT;
}

static int cmd_helo(char *arg, State * const state)
{
   if (state->helo != NULL) {
      say_flush(state, _("503 Please choose between HELO and EHLO and say it only once") CRLF);
      return CMD_RETURN_INVALID;
   }
   if (arg == NULL) {
      say_flush(state, _("501 Missing hostname") CRLF);
      return CMD_RETURN_INVALID;
   }   
   STR2_DUP(state->helo, arg);
   if (state->helo == NULL) {
      say_flush(state, _("421 Out of memory") CRLF);
      return CMD_RETURN_FAILURE;
   }
   say_flush(state, _("250 hi") CRLF);
      
   return 0;
}

static int cmd_size(char *arg, State * const state)
{
   unsigned long wanted_size = 0UL;
   
   if (state->helo == NULL) {
      say_flush(state, _("503 Say EHLO first and I'll tell you then") CRLF);
      return CMD_RETURN_INVALID;
   }   
   if (arg != NULL && isdigit(*arg) != 0) {
      wanted_size = strtoul(arg, NULL, 0);
   } 
   if (wanted_size == 0UL) {
      say_flush(state, _("503 Ask properly and I'll tell you then") CRLF);
      return CMD_RETURN_INVALID;      
   }
   if (wanted_size < config.databytes) {
      say_flush(state, _("220 ok") CRLF);
   } else {
      say_flush(state, _("503 too much") CRLF);      
   }
   
   return 0;
}

static int cmd_ehlo(char *arg, State * const state)
{
   if (state->helo != NULL) {
      say_flush(state, _("503 Please choose between HELO and EHLO") CRLF);
      return CMD_RETURN_INVALID;
   }   
   if (arg == NULL) {
      say_flush(state, _("501 Missing hostname") CRLF);
      return CMD_RETURN_INVALID;
   }  
   STR2_DUP(state->helo, arg);
   if (state->helo == NULL) {
      say_flush(state, _("421 Out of memory") CRLF);
      return CMD_RETURN_FAILURE;  
   }
   say_flush(state, _("250-hi") CRLF
	     _("250-8BITMIME") CRLF _("250-PIPELINING") CRLF _("250 SIZE") CRLF);

   return 0;
}

static int cmd_noop(char *arg, State * const state)
{
   (void) arg;
   (void) state;
   say_flush(state, _("250 ZzzZzzz") CRLF);
   
   return 0;
}

static int cmd_vrfy(char *arg, State * const state)
{
   (void) state;
   
   if (arg == NULL) {
      say_flush(state, _("501 Where is the user name ?") CRLF);
      return CMD_RETURN_INVALID;
   }   
   say_flush(state, _("252 Gimme mail and we'll see") CRLF);
   
   return 0;
}

static int cmd_expn(char *arg, State * const state)
{
   (void) state;
   
   if (arg == NULL) {
      say_flush(state, _("501 Where is the address ?") CRLF);
      return CMD_RETURN_INVALID;
   }   
   say_flush(state, _("502 Expand it yourself") CRLF);
   
   return 0;
}

void free_state(State * const state, int keep_some)
{
   if (state->to != NULL) {
      list_destroy(state->to);
      free(state->to);
      state->to = NULL;      
   }
   if (state->fromto != NULL) {
      var_string_free(state->fromto);
      state->fromto = NULL;
   }
   if (state->from != NULL) {
      free(state->from);
      state->from = NULL;
   }
   if (keep_some == 0) {
      if (state->helo != NULL) {
	 free(state->helo);
	 state->helo = NULL;
      }
   }
   if (state->data_buf != NULL) {
      free(state->data_buf);
      state->data_buf = NULL;
   }
   state->data_buf_pnt = NULL;
   if (state->data_buf_size > (size_t) 0U) {
      state->data_buf_size = (size_t) 0U;
   }
   if (state->qmail_queue.mess_filename != NULL) {
      free(state->qmail_queue.mess_filename);
      state->qmail_queue.mess_filename = NULL;
   }
#ifdef INTD_COMPATIBILITY
   if (state->qmail_queue.intd_filename != NULL) {
      free(state->qmail_queue.intd_filename);
      state->qmail_queue.intd_filename = NULL;
   }
#endif
}

void init_state(State * const state, int keep_some)
{
   state->to = NULL;
   state->fromto = NULL;
   state->from = NULL;
   if (keep_some == 0) {
      state->helo = NULL;
   }
   state->data_buf = NULL;
   state->data_buf_pnt = NULL;
   state->data_buf_size = (size_t) 0U;
   state->data_size = (size_t) 0U;
   state->eof_status = DATA_INITIAL_EOF_STATUS;
   state->header_status = DATA_INITIAL_HEADER_STATUS;
   state->header_received_pnt = HEADER_RECEIVED;
   state->header_delivered_pnt = HEADER_DELIVERED;
   state->hops = 0;
   state->qmail_queue.mess_filename = NULL;
#ifdef INTD_COMPATIBILITY
   state->qmail_queue.intd_filename = NULL;
#endif
   state->qmail_queue.inode = 0UL;
   state->qmail_queue.mess_fd = -1;
#ifdef INTD_COMPATIBILITY
   state->qmail_queue.intd_fd = -1;
#endif
}

static int cmd_rset(char *arg, State * const state)
{
   (void) arg;
   free_state(state, 1);
   init_state(state, 1);
   say_flush(state, _("250 Done") CRLF);
   
   return 0;
}

static int cmd_mail_parse(State * const state, 
			  char *arg, const char *arg2, char **dst_)
{
   char *dst = NULL, *dstpnt = NULL;
   char has_brackets = 0;
   char seen_backslash = 0, seen_quote = 0;

   *dst_ = NULL;
   if (arg == NULL) {
      say_flush(state, _("500 Sorry, but... no") CRLF);
      return CMD_RETURN_INVALID;
   }
   if (strncasecmp(arg, arg2, strlen(arg2)) == 0) {
      char c = *(arg + strlen(arg2));
      
      if (c == ' ' || c == ':') {
	 arg += strlen(arg2);
	 arg = skip_spaces(arg);
      }
   }
   if (*arg == 0) {
      say_flush(state, _("501 Sorry, but I need a correct address") CRLF);
      return CMD_RETURN_INVALID;
   }
   if (*arg == ':') {
      arg = skip_spaces(arg + 1);
   }
   if (*arg == '<') {
      arg++;
      has_brackets = 1;
   }
   if (*arg == 0) {
      say_flush(state, _("501 So where is the address ?") CRLF);
      return CMD_RETURN_INVALID;
   }      
   if (*arg == '@') {
      while (*arg != 0 && *arg++ != ':');
      if (*arg == 0) {
	 say_flush(state, _("501 It is only a source route") CRLF);
	 return CMD_RETURN_INVALID;
      }
   }
   if ((dst = malloc(strlen(arg) + 1)) == NULL) {
      say_flush(state, _("421 Out of memory") CRLF);
      return CMD_RETURN_INVALID;
   }
   dstpnt = dst;
   while (*arg != 0) {
      if (seen_backslash != 0) {
	 *dstpnt++ = *arg;
	 seen_backslash = 0;
      } else {
	 if (seen_quote == 0) {
	    if (has_brackets == 0) {
	       if (*arg == ' ') {
		  break;
	       }
	    } else if (*arg == '>') {
	       break;
	    }
	 }
	 if (*arg == '\"') {
	    seen_quote ^= 1;
	 } else if (*arg == '\\') {
	    seen_backslash = 1;
	 } else {
	    *dstpnt++ = *arg;
	 }
      }
      arg++;
   }
   *dstpnt = 0;      
   *dst_ = dst;
   
   return 0;
}

static int cmd_mail(char *arg, State * const state)
{   
   char *dst;
   int ret;
   
   if (state->helo == NULL) {
      say_flush(state, _("503 Polite people say hello first") CRLF);
      return CMD_RETURN_INVALID;
   }
   if (state->from != NULL) {
      say_flush(state, _("503 We already have a sender here") CRLF);
      return CMD_RETURN_INVALID;
   }
   if ((ret = cmd_mail_parse(state, arg, "FROM", &dst)) != 0) {
      return ret;
   }
   if (dst == NULL || *dst == 0) {
      say_flush(state, _("503 Null senders are not allowed") CRLF);
      free(dst);
      return CMD_RETURN_INVALID;     
   }
   if (check_email_consistency(dst) <= 0) {
      say_flush(state, _("553 Sorry, this doesn't look like a valid email address") CRLF);
      free(dst);
      return CMD_RETURN_INVALID;      
   }
   if (check_badmailfrom(&config, dst) > 0) {
      say_flush(state, _("553 Sorry, your address is banned from here") CRLF);
      free(dst);
      return CMD_RETURN_INVALID;
   }
   state->from = dst;
#ifdef DEBUG
   say(state, _("250 So, your address is <"));
   say(state, _(dst));
   say_flush(state, ">" CRLF);
#else
   say_flush(state, _("250 ok") CRLF);
#endif
   
   return 0;
}

static int cmd_rcpt(char *arg, State * const state)
{   
   char *dst = NULL;
   int ret;
   Node *new_node = NULL;
   int exit_return;
   
   if (state->helo == NULL) {
      say_flush(state, _("503 Polite people say hello first") CRLF);
      exit_return = CMD_RETURN_INVALID;
      goto exit_failure;
   }
   if (state->from == NULL) {
      say_flush(state, _("503 Sorry, but who are you ?") CRLF);
      exit_return = CMD_RETURN_INVALID;
      goto exit_failure;
   }
   if ((ret = cmd_mail_parse(state, arg, "TO", &dst)) != 0) {
      say_flush(state, _("503 Invalid recipient") CRLF);
      exit_return = ret;
      goto exit_failure;
   }
   if (dst == NULL || *dst == 0) {
      say_flush(state, _("503 Null recipients are not allowed") CRLF);
      exit_return = CMD_RETURN_INVALID;      
      goto exit_failure;
   }
   if (check_email_consistency(dst) <= 0) {
      say_flush(state, _("553 Sorry, this doesn't look like a valid email address") CRLF);
      exit_return = CMD_RETURN_INVALID;      
      goto exit_failure;      
   }
   if (config.open_relay == 0) {
      if (check_rcpthosts(&config, dst) == CHECK_RCPTHOSTS_RETURN_OK) {
	 goto relaying_ok;
      }
   } else if (config.morercpthosts.fd < 0) {
      goto relaying_ok;
   }
   if (config.morercpthosts.fd >= 0 &&
       check_morercpthosts(&config, dst) == CHECK_RCPTHOSTS_RETURN_OK) {
      goto relaying_ok;
   }
   say_flush(state, _("553 I don't relay that domain") CRLF);
   exit_return = CMD_RETURN_INVALID;      
   goto exit_failure;
   
   relaying_ok:
   if ((new_node = malloc(sizeof *new_node)) == NULL) {
      say_flush(state, _("421 Out of memory") CRLF);
      exit_return = CMD_RETURN_FAILURE;
      goto exit_failure;
   }
   new_node->next = NULL;
   STR2_DUP(new_node->object, dst);
   if (new_node->object == NULL) {
      free(new_node);
      say_flush(state, _("421 Out of memory") CRLF);
      exit_return = CMD_RETURN_FAILURE;      
      goto exit_failure;
   }
   if (state->to == NULL) {
      if ((state->to = malloc(sizeof *state->to)) == NULL) {
	 free(new_node->object);
	 new_node->object = NULL;
	 say_flush(state, _("421 Out of memory") CRLF);
	 exit_return = CMD_RETURN_FAILURE;
	 goto exit_failure;
      }      
      state->to->first = state->to->last = new_node;
   } else {      
      state->to->last->next = new_node;
      state->to->last = new_node;
   }
#ifdef DEBUG
   say(state, _("250 Okay, sending to <"));
   say(state, _(dst));
   say_flush(state, ">" CRLF);
#else
   say_flush(state, _("250 ok") CRLF);
#endif
   
   if (dst != NULL) {
      free(dst);
   }   
   return 0;
   
   exit_failure:
   if (dst != NULL) {
      free(dst);
   }
   if (new_node != NULL) {
      free(new_node);
   }
   
   return exit_return;
}

int erase_queue(QmailQueue * const qmail_queue)
{
#ifdef INTD_COMPATIBILITY
   if (qmail_queue == NULL ||
       qmail_queue->mess_filename == NULL ||
       qmail_queue->mess_filename_size < 3U ||
       qmail_queue->intd_filename == NULL ||
       qmail_queue->intd_filename_size < 3U) {
      D(puts("Invalid qmail queue pointer"));
      return -1;
   }
#else
   if (qmail_queue == NULL ||
       qmail_queue->mess_filename == NULL ||
       qmail_queue->mess_filename_size < 3U) {
      D(puts("Invalid qmail queue pointer"));
      return -1;
   }   
#endif
   if (qmail_queue->mess_fd >= 0) {
      close(qmail_queue->mess_fd);
      qmail_queue->mess_fd = -1;
   }
#ifdef INTD_COMPATIBILITY
   if (qmail_queue->intd_fd >= 0) {
      close(qmail_queue->intd_fd);
      qmail_queue->intd_fd = -1;
   }
#endif
   unlink(qmail_queue->mess_filename);
#ifdef INTD_COMPATIBILITY
   unlink(qmail_queue->intd_filename);
#endif
   
   return 0;
}

static inline int data_bufferize(const int c, State * const state)
{   
   *(state->data_buf_pnt++) = c;
   if ((size_t) (state->data_buf_pnt - state->data_buf) >=
       state->data_buf_size) {
      state->data_buf_pnt = NULL;
      D(puts("> Writing data chunk"));
      if (safe_write(state->qmail_queue.mess_fd, 
		     state->data_buf, state->data_buf_size) < 0) {
	 D(perror("data_buf safe_write"));
	 return -1;
      }      
      state->data_buf_pnt = state->data_buf;
   }
   
   return 0;
}

static int data_bufferize_flush(State * const state)
{
   size_t data_buf_size = 
     (size_t) (state->data_buf_pnt - state->data_buf);
   char *todo_filename = NULL;
   size_t todo_filename_size;

   state->data_buf_pnt = NULL;
   D(puts("Writing the final chunk"));
   D(printf("mess_fd = %d\n" __ state->qmail_queue.mess_fd));
   if (safe_write(state->qmail_queue.mess_fd, 
		  state->data_buf, data_buf_size) < 0) {
      D(perror("data_buf safe_write"));
      goto exit_failure;
   }     
#ifndef LAZY_METADATA_SYNC
# ifdef INTD_COMPATIBILITY
   if (state->qmail_queue.intd_fd >= 0) {
      fsync(state->qmail_queue.intd_fd);
      close(state->qmail_queue.intd_fd);
      state->qmail_queue.intd_fd = -1;
   }      
# endif
   if (state->qmail_queue.mess_fd >= 0) {
      fsync(state->qmail_queue.mess_fd);
      close(state->qmail_queue.mess_fd);      
      state->qmail_queue.mess_fd = -1;  
   }
#endif
#ifdef INTD_COMPATIBILITY
   todo_filename_size = (sizeof QUEUE_PATH_TODO - 1) + 1 +
     size_unsigned_long(state->qmail_queue.inode) + 1;
#else
   todo_filename_size = (sizeof QUEUE_PATH_REISERTODO - 1) + 1 +
     size_unsigned_long(state->qmail_queue.inode) + 1;     
#endif
   if ((todo_filename = alloca(todo_filename_size)) == NULL) {
      D(perror("alloca"));
      goto exit_failure;
   }
#ifdef INTD_COMPATIBILITY
      {
	 register char *todo_filename_strpnt;
	 
	 STR2_INIT(todo_filename, todo_filename_strpnt);
	 STR2_UNSAFE_STRCAT_L(todo_filename_strpnt, QUEUE_PATH_TODO "/", sizeof (QUEUE_PATH_TODO "/"));
	 STR2_UNSAFE_ULINTCAT(todo_filename_strpnt, state->qmail_queue.inode);
	 STR2_UNSAFE_FINALIZE(todo_filename_strpnt);
      }
#else
      {
	 register char *todo_filename_strpnt;
	 
	 STR2_INIT(todo_filename, todo_filename_strpnt);
	 STR2_UNSAFE_STRCAT_L(todo_filename_strpnt, QUEUE_PATH_REISERTODO "/", sizeof (QUEUE_PATH_REISERTODO "/"));
	 STR2_UNSAFE_ULINTCAT(todo_filename_strpnt, state->qmail_queue.inode);
	 STR2_UNSAFE_FINALIZE(todo_filename_strpnt);
      }
#endif
#ifdef INTD_COMPATIBILITY   
   D(printf("Linking from [%s] to [%s]\n" __
	    state->qmail_queue.intd_filename __ todo_filename));
   if (link(state->qmail_queue.intd_filename, todo_filename) < 0) {
      D(perror("link"));
      goto exit_failure;
   }
     {
	int link_sync;
	
	if ((link_sync = open(todo_filename, O_RDONLY | O_NDELAY)) < 0) {
	   goto exit_failure;
	}
	fsync(link_sync);
	close(link_sync);
     }
#else
   D(printf("Linking from [%s] to [%s]\n" __
	    state->qmail_queue.mess_filename __ todo_filename));
   if (link(state->qmail_queue.mess_filename, todo_filename) < 0) {
      D(perror("link"));
      goto exit_failure;
   }   
#endif
   ALLOCA_FREE(todo_filename);
   todo_filename = NULL;

#ifndef LAZY_METADATA_SYNC
# ifdef INTD_COMPATIBILITY
   trigger_qmail_send();  
# else
   trigger_lmail_send();  
# endif
#else
      {
	struct sembuf sb;
	
	sb.sem_num = 0;
	sb.sem_op = 1;
	sb.sem_flg = 0;
	if (semop(sync_semid, &sb, 1) < 0) {
	   D(perror("semop"));
	   goto exit_failure;
	}	
	if (sync_queue_push(&global_sync_queue,
# ifdef INTD_COMPATIBILITY
			    state->qmail_queue.intd_fd,
# endif
			    state->qmail_queue.mess_fd) < 0) {
	   puts("sync_queue_push wrong args");
	   goto exit_failure;
	}
	 sb.sem_num = 0;
	 sb.sem_op = 0;
	 sb.sem_flg = 0;
# ifdef INTD_COMPATIBILITY
	 D(printf("Waiting for semaphore release, fd = %d,%d\n" __
		  state->qmail_queue.intd_fd __
		  state->qmail_queue.mess_fd));
# else
	D(printf("Waiting for semaphore release, fd = %d\n" __
		 state->qmail_queue.mess_fd));
# endif
	 if (semop(sync_semid, &sb, 1) < 0) {
	    D(perror("semop"));
# ifdef INTD_COMPATIBILITY
	    fsync(state->qmail_queue.intd_fd);
	    close(state->qmail_queue.intd_fd);
	    state->qmail_queue.intd_fd = -1;
# endif
	    fsync(state->qmail_queue.mess_fd);
	    close(state->qmail_queue.mess_fd);
	    state->qmail_queue.mess_fd = -1;	   	   
	 } 
# ifdef INTD_COMPATIBILITY
	D(printf("Received semaphore release, fd = %d,%d\n" __
		 state->qmail_queue.intd_fd __
		 state->qmail_queue.mess_fd));
# else
	D(printf("Received semaphore release, fd = %d\n" __
		 state->qmail_queue.mess_fd));	 
# endif
     }
#endif
         
   return 0;
   
   exit_failure :
#ifdef INTD_COMPATIBILITY
   if (state->qmail_queue.intd_fd >= 0) {
      fsync(state->qmail_queue.intd_fd);
      close(state->qmail_queue.intd_fd);
      state->qmail_queue.intd_fd = -1;
   }
#endif
   if (state->qmail_queue.mess_fd >= 0) {
      fsync(state->qmail_queue.mess_fd);
      close(state->qmail_queue.mess_fd);      
      state->qmail_queue.mess_fd = -1;  
   }
#ifdef INTD_COMPATIBILITY
   if (todo_filename != NULL) {
      ALLOCA_FREE(todo_filename);
   }
#endif
   erase_queue(&state->qmail_queue);
   
   return -1;
}

int fromto_write(State *state,
		 const char * const fromto_formatted,
		 const size_t fromto_size)
{
   const QmailQueue *qmail_queue;
   char *buf = NULL, *buf_pnt;
   size_t s;
   size_t uid_size, pid_size;
   unsigned long pid;
   unsigned long uid;
   int exit_return = 0;
   
   if (state == NULL || (qmail_queue = &state->qmail_queue) == NULL) {
      D(puts("Qmail queue in fromto_write is invalid"));
      goto exit_failure;
   }
#ifdef INTD_COMPATIBILITY
   if (qmail_queue->intd_fd < 0) {
      D(puts("Qmail queue in fromto_write has a null descriptor"));
      goto exit_failure;
   }
#else
   if (qmail_queue->mess_fd < 0) {
      D(puts("Qmail queue in fromto_write has a null descriptor"));
      goto exit_failure;
   }
#endif
   pid = (unsigned long) getpid();
   uid = (unsigned long) getuid();
   s = 1U;			       /* u */
   s += (uid_size = 
	 size_unsigned_long(uid));     /* 42 */
   s++;				       /* \0 */
   s++;				       /* p */
   s += (pid_size =
	 size_unsigned_long(pid));     /* 42 */
   s++;				       /* \0 */
   s += fromto_size;		       /* F...T...T... */
   if ((buf = alloca(s + 1)) == NULL) {
      D(perror("malloc"));
      goto exit_failure;
   }   
   STR2_INIT(buf, buf_pnt);
   STR2_UNSAFE_CHARCAT(buf_pnt, 'u');
   STR2_UNSAFE_ULINTCAT(buf_pnt, uid);
   STR2_UNSAFE_CHARCAT(buf_pnt, 0);
   STR2_UNSAFE_CHARCAT(buf_pnt, 'p');
   STR2_UNSAFE_ULINTCAT(buf_pnt, pid);
   STR2_UNSAFE_CHARCAT(buf_pnt, 0);
   memcpy(buf_pnt, fromto_formatted, fromto_size);
#ifdef INTD_COMPATIBILITY
   if (safe_write(qmail_queue->intd_fd, buf, s) < 0) {
      D(perror("fromto safe_write"));
      goto exit_failure;
   } 
#else
      {
	 size_t sc = s;
	 buf_pnt = buf;
	 
	 do {
	    if (++(state->data_size) > config.databytes) {
	       D(puts("size exceeded"));
	       say_flush(state, _("552 Message too long") CRLF);
	       goto exit_failure;
	    }	 
	    if (data_bufferize(*buf_pnt, state) < 0) {
	       D(perror("fromto data_bufferize"));
	       say_flush(state, _("421 I/O error") CRLF);
	       goto exit_failure;
	    }
	    buf_pnt++;
	    sc--;
	 } while (sc != (size_t) 0U);
	 	 if (state->fromto->size > 0xffffffff) {
	    say_flush(state, _("421 Envelope too big") CRLF);
	    exit_return = CMD_RETURN_FAILURE;
	    goto exit_failure;	    	    
	 }
	 if (data_bufferize((s >> 24) & 0xff, state) < 0 ||
	     data_bufferize((s >> 16) & 0xff, state) < 0 ||
	     data_bufferize((s >> 8) & 0xff, state) < 0 ||
	     data_bufferize(s & 0xff, state) < 0) {
	    D(perror("Unable to write fromto->size"));
	    exit_return = CMD_RETURN_FAILURE;	    
	    goto exit_failure;	    	    
	 }
      }
#endif
   goto exit_normal;
   
   exit_failure:
   exit_return = -1;
   
   exit_normal:
   if (buf != NULL) {
      ALLOCA_FREE(buf);
   }
   
   return exit_return;
}

static inline int cmd_data_fetch(const int c, State * const state)
{
   int exit_return = 0;
   
   if (state->header_received_pnt != NULL) {
      if (state->header_status == 0 && c == '\r') {
	 state->header_status = 1;
      } else if (state->header_status == 1 && c == '\n') {
	 state->header_status = 2;
      } else if (state->header_status == 2 && c == '\r') {
	 state->header_status = 3;
      } else if (state->header_status == 3 && c == '\n') {
	 state->header_status = DATA_INITIAL_HEADER_STATUS;
	 state->header_received_pnt = NULL;
	 state->header_delivered_pnt = NULL;
	 D(puts("End of data header"));
      } else {
	 state->header_status = 0;
      }
   }   
   switch (state->eof_status) {            
    case 0 :
      if (c == '\r') {
	 state->eof_status = 4;
	 return 0;
      } 
      break;
    case 1 :
      if (c == '.') {
	 state->eof_status = 2;
	 return 0;
      } else if (c == '\r') {
	 state->eof_status = 4;
	 return 0;
      }
      state->eof_status = 0;
      break;
    case 2 :
      if (c == '\r') {
	 state->eof_status = 3;
	 return 0;
      }
      state->eof_status = 0;
      break;
    case 3 :
      if (c == '\n') {
#ifndef INTD_COMPATIBILITY
	 if (data_bufferize(0, state) < 0 ||
	     data_bufferize(0, state) < 0) {
	    D(perror("Can't add delivery info to a message"));
	    say_flush(state, _("421 Can't write delivery info") CRLF);
	    exit_return = CMD_RETURN_FAILURE;
	    goto exit_failure;	    
	 }
#endif
	 if (fromto_write(state, state->fromto->str, state->fromto->size) < 0) {
	    D(perror("Failure with fromto_write"));
	    say_flush(state, _("421 Can't write fromto") CRLF);
	    exit_return = CMD_RETURN_FAILURE;
	    goto exit_failure;	    
	 }
#ifndef INTD_COMPATIBILITY
	    {
	       size_t message_footer_cnt = (size_t) 0U;
	       
	       do {
		  if (data_bufferize(MESSAGE_FOOTER[message_footer_cnt], 
				     state) < 0) {
		     D(perror("Unable to write the message footer"));
		     say_flush(state, _("421 Can't write message footer") CRLF);
		     exit_return = CMD_RETURN_FAILURE;
		     goto exit_failure;	    	    		     
		  }		  
		  message_footer_cnt++;
	       } while (message_footer_cnt < (sizeof MESSAGE_FOOTER - 1));
	       
	    }
#endif
	 if (data_bufferize_flush(state) < 0) {
	    exit_return = CMD_RETURN_FAILURE;
	    goto exit_failure;
	 }
	 if (state->data_buf != NULL) {
	    free(state->data_buf);
	    state->data_buf = NULL;
	 }
	 state->data_buf_pnt = NULL;
	 say_flush(state, _("250 qd") CRLF);
	 
	 return 0;
      }
      if (++(state->data_size) > config.databytes) {
	 D(puts("size exceeded"));
	 say_flush(state, _("552 Message too long") CRLF);
	 exit_return = CMD_RETURN_FAILURE;
	 goto exit_failure;
      }	       
      if (data_bufferize('.', state) < 0 ||
	  data_bufferize('\r', state) < 0) {
	 exit_return = CMD_RETURN_FAILURE;	 
	 goto exit_failure;
      }
      if (c == '\r') {
	 state->eof_status = 4;
	 return 0;
      }
      state->eof_status = 0;
      break;
    case 4 :
      if (c == '\n') {
	 if (state->header_received_pnt != NULL) {
	    state->header_received_pnt = HEADER_RECEIVED;
	 }
	 if (state->header_delivered_pnt != NULL) {
	    state->header_delivered_pnt = HEADER_DELIVERED;
	 }
	 state->eof_status = 1;
      } else if (c == '\r') {
	 if (++(state->data_size) > config.databytes) {
	    D(puts("size exceeded"));
	    say_flush(state, _("552 Message too long") CRLF);
	    exit_return = CMD_RETURN_FAILURE;
	    goto exit_failure;
	 }	       	 
	 if (data_bufferize('.', state) < 0) {
	    exit_return = CMD_RETURN_FAILURE;	    
	    goto exit_failure;
	 }
	 state->eof_status = 0;
      }
   }
   if (state->header_received_pnt != NULL && 
       tolower(c) == *(state->header_received_pnt)) {
      state->header_received_pnt++;
      if (*(state->header_received_pnt) == 0) {
	 state->hops++;
	 if (state->hops > config.max_hops) {
	    say_flush(state, _("554 Too many hops") CRLF);
	    exit_return = CMD_RETURN_TOO_MANY_HOPS;
	    goto exit_failure;
	 } else {
	    D(printf("Hops = %u\n" __ state->hops));
	 }      	 
      }
   }
   if (state->header_delivered_pnt != NULL && 
       tolower(c) == *(state->header_delivered_pnt)) {
      state->header_delivered_pnt++;
      if (*(state->header_delivered_pnt) == 0) {
	 state->hops++;
	 if (state->hops > config.max_hops) {
	    say_flush(state, _("554 Too many hops") CRLF);	    
	    exit_return = CMD_RETURN_TOO_MANY_HOPS;
	    goto exit_failure;
	 } else {
	    D(printf("Hops = %u\n" __ state->hops));
	 }      	 
      }
   }
   if (++(state->data_size) > config.databytes) {
      D(puts("size exceeded"));
      say_flush(state, _("552 Message too long") CRLF);
      exit_return = CMD_RETURN_FAILURE;
      goto exit_failure;
   }	       
   if (data_bufferize(c, state) < 0) {
      exit_return = CMD_RETURN_FAILURE;	    
      goto exit_failure;      
   }   
   return 0;
   
   exit_failure:
   erase_queue(&state->qmail_queue);
   if (state->data_buf != NULL) {
      free(state->data_buf);
      state->data_buf = 0;
   }
   return exit_return;
}

int open_queue(QmailQueue * const qmail_queue)
{   
   unsigned long sequence;
   char *mess_filename;
   size_t mess_filename_size;
#ifdef INTD_COMPATIBILITY
   char *intd_filename;
   size_t intd_filename_size;
#endif
   int mess_fd = -1;
#ifdef INTD_COMPATIBILITY
   int intd_fd = -1;
#endif
   
#ifdef INTD_COMPATIBILITY
   if (qmail_queue == NULL ||
       (mess_filename = qmail_queue->mess_filename) == NULL ||
       (mess_filename_size = qmail_queue->mess_filename_size) < 3U ||
       (intd_filename = qmail_queue->intd_filename) == NULL ||
       (intd_filename_size = qmail_queue->intd_filename_size) < 3U) {
      D(puts("Invalid qmail queue pointer"));
      return -1;
   }
#else
   if (qmail_queue == NULL ||
       (mess_filename = qmail_queue->mess_filename) == NULL ||
       (mess_filename_size = qmail_queue->mess_filename_size) < 3U) {
      D(puts("Invalid qmail queue pointer"));
      return -1;
   }   
#endif
   qmail_queue->mess_fd = - 1;
#ifdef INTD_COMPATIBILITY
   qmail_queue->intd_fd = -1;
#endif
   qmail_queue->inode = 0UL;
   for (;;) {
      global_sequence++;
      sequence = global_sequence;
      D(printf("Sequence number : [%lu]\n" __ sequence));      
      qmail_queue->inode = sequence;
#ifdef INTD_COMPATIBILITY
      {
	 register char *intd_filename_strpnt;
	 
	 STR2_INIT(intd_filename, intd_filename_strpnt);
	 STR2_UNSAFE_STRCAT_L(intd_filename_strpnt, QUEUE_PATH_INTD "/", sizeof (QUEUE_PATH_INTD "/"));
	 STR2_UNSAFE_ULINTCAT(intd_filename_strpnt, sequence);
	 STR2_UNSAFE_FINALIZE(intd_filename_strpnt);
      }
      D(printf("Intd file name : [%s]\n" __ intd_filename));      
      if ((intd_fd = open(intd_filename, O_WRONLY | O_CREAT | O_EXCL |
			  O_NDELAY, 0660)) < 0) {
	 continue;
      }
#endif
#ifdef INTD_COMPATIBILITY
	 {
	    register char *mess_filename_strpnt;
	    
	    STR2_INIT(mess_filename, mess_filename_strpnt);
	    STR2_UNSAFE_STRCAT_L(mess_filename_strpnt, QUEUE_PATH_MESS "/", sizeof (QUEUE_PATH_MESS "/"));
	    STR2_UNSAFE_ULINTCAT(mess_filename_strpnt, sequence % (unsigned long) QMAIL_MESS_SPLIT);
	    STR2_UNSAFE_CHARCAT(mess_filename_strpnt, '/');
	    STR2_UNSAFE_ULINTCAT(mess_filename_strpnt, sequence);
	    STR2_UNSAFE_FINALIZE(mess_filename_strpnt);
	 }
#else
	 {
	    register char *mess_filename_strpnt;
	    
	    STR2_INIT(mess_filename, mess_filename_strpnt);
	    STR2_UNSAFE_STRCAT_L(mess_filename_strpnt, QUEUE_PATH_REISERMESS "/", sizeof (QUEUE_PATH_REISERMESS "/"));
	    STR2_UNSAFE_ULINTCAT(mess_filename_strpnt, sequence % (unsigned long) QMAIL_REISERMESS_SPLIT);
	    STR2_UNSAFE_CHARCAT(mess_filename_strpnt, '/');
	    STR2_UNSAFE_ULINTCAT(mess_filename_strpnt, sequence);
	    STR2_UNSAFE_FINALIZE(mess_filename_strpnt);
	 }
#endif
#ifndef PREALLOCATE
      mess_fd = open(mess_filename, O_WRONLY | O_CREAT | O_EXCL | O_NDELAY,
		     0660);
#else
	 {
	    struct stat st;
	    
	    if (stat(mess_filename, &st) < 0) {
	       mess_fd = open(mess_filename, O_WRONLY | O_CREAT | O_EXCL | O_NDELAY,
			      0660);
	    } else {
	       mess_fd = open(mess_filename, O_WRONLY | O_NDELAY,
			      0660);	       	       
	    }
	 }
#endif
      if (mess_fd  < 0) {
# ifdef INTD_COMPATIBILITY
	 close(intd_fd);
	 unlink(intd_filename);
# endif
	 continue;
      }      
      break;
   }
   D(printf("Mess file name : [%s] fd = [%d]\n" __ mess_filename __ mess_fd));
#ifdef INTD_COMPATIBILITY
   qmail_queue->intd_fd = intd_fd;
#endif
   qmail_queue->mess_fd = mess_fd;   
   
   return 0;
}

static int cmd_data_check_extensions(char *arg) 
{
   unsigned long wanted_size;
   char *arg_next;
   int exit_return = 0;
   
   if (arg == NULL) {      
      return 0;
   }
   for (;;) {
      arg = skip_spaces(arg);
      if (*arg == 0) {
	 return exit_return;
      }
      if (strncasecmp(arg, EXTENSION_DATA_SIZE, (sizeof EXTENSION_DATA_SIZE) - 1) == 0) {
	 arg = skip_spaces(arg + (sizeof EXTENSION_DATA_SIZE) - 1);
	 if (*arg != '=') {
	    ext_error:
	    D(puts("Syntax error with " EXTENSION_DATA_SIZE " extension"));
	    goto continue_loop;
	 }
	 arg = skip_spaces(arg + 1);
	 if (isdigit(*arg) == 0) {
	    goto ext_error;
	 }
	 wanted_size = strtoul(arg, &arg_next, 0);
	 if (arg_next != NULL) {
	    arg = arg_next;
	 }
	 if (wanted_size == 0UL) {
	    goto ext_error;
	 }	 
	 D(printf("Client is annoucing the following size : [%lu]\n",
		  wanted_size));
	 if (wanted_size < config.databytes) {
	    D(puts("Size ok"));
	    exit_return = 0;
	    goto continue_loop;
	 } else {
	    D(puts("Message too big"));
	    exit_return = -1;
	    goto continue_loop;
	 }
      }
      continue_loop:
      if (*arg != 0) {
	 arg++;
      } else {
	 break;
      }
   }   
   
   return exit_return;
}

static int cmd_data(char *arg, State * const state)
{   
   VarString *received = NULL;
   char *received_pnt;
   Node *node_pnt;
   int exit_return = 0;
   
   if (state->helo == NULL || state->from == NULL || state->to == NULL) {
      say_flush(state, _("503 I don't speak to unknown people") CRLF);
      exit_return = CMD_RETURN_INVALID;      
      goto exit_failure;
   }
   if (cmd_data_check_extensions(arg) < 0) {
      say_flush(state, _("503 Too much") CRLF);
      exit_return = CMD_RETURN_INVALID;      
      goto exit_failure;
   }
   say_flush(state, _("354 go") CRLF);
   if (state->to == NULL || state->to->first == NULL ||
       (state->fromto = var_string_new()) == NULL ||
       var_string_add(state->fromto, "F") != 0 ||
       var_string_cat(state->fromto, state->from) != 0) {
      say_flush(state, _("421 Out of memory") CRLF);
      exit_return = CMD_RETURN_FAILURE;
      goto exit_failure;
   }
   node_pnt = state->to->first;
   while (node_pnt != NULL) {
      if (var_string_add(state->fromto, "T") != 0 ||
	  var_string_cat(state->fromto, (char *) node_pnt->object) != 0) {
	 say_flush(state, _("421 Out of memory") CRLF);
	 exit_return = CMD_RETURN_FAILURE;
	 goto exit_failure;
      }	   
      node_pnt = node_pnt->next;
   }
   
   
     {	
#ifdef INTD_COMPATIBILITY
	state->qmail_queue.mess_filename_size =
	  (sizeof QUEUE_PATH_MESS - 1U) + 1U /* / */ + 
	  (STRSIZE_QMAIL_MESS_SPLIT - 1U) + 1U /* / */ + STRSIZE_ULONG + 1U;
#else
	state->qmail_queue.mess_filename_size =
	  (sizeof QUEUE_PATH_REISERMESS - 1U) + 1U /* / */ +
	  (STRSIZE_QMAIL_REISERMESS_SPLIT - 1U) + 1U /* / */ + STRSIZE_ULONG + 1U;
#endif
#ifdef INTD_COMPATIBILITY
	state->qmail_queue.intd_filename_size =
	  (sizeof QUEUE_PATH_INTD - 1U) + 1U /* / */ + STRSIZE_ULONG + 1U;
#endif
	
	if ((state->qmail_queue.mess_filename = 
	     malloc(state->qmail_queue.mess_filename_size)) == NULL
#ifdef INTD_COMPATIBILITY
	    || (state->qmail_queue.intd_filename =
	     malloc(state->qmail_queue.intd_filename_size)) == NULL
#endif
	    ) {
	   D(perror("malloc"));
	   say_flush(state, _("421 Out of memory") CRLF);
	   exit_return = CMD_RETURN_FAILURE;
	   goto exit_failure;
	}
	
	D(puts("> Opening the queue"));
	if (open_queue(&state->qmail_queue) != 0) {
	   perror("> Unable to open the queue");
	   say_flush(state, _("421 Unable to open the queue") CRLF);
	   exit_return = CMD_RETURN_FAILURE;
	   goto exit_failure;
	}
#ifdef INTD_COMPATIBILITY
	if (fromto_write(state, state->fromto->str, state->fromto->size) < 0) {
	   D(perror("> Failure with fromto_write"));
	   say_flush(state, _("421 Can't write fromto") CRLF);
	   exit_return = CMD_RETURN_FAILURE;
	   goto exit_failure;
	}		
#endif
     }
               
   if ((state->data_buf = malloc(DATA_CHUNK_SIZE)) == NULL) {
      D(perror("malloc"));
      say_flush(state, _("421 Out of memory") CRLF);
      exit_return = CMD_RETURN_FAILURE;
      goto exit_failure;
   }
   if ((received = rfc_received(getenv("TCPREMOTEHOST"),
				state->helo,
				getenv("TCPREMOTEINFO"),
				getenv("TCPREMOTEIP"),
				getenv("TCPLOCALHOST"),
				"SMTP")) == NULL) {
      say_flush(state, _("421 Unable to create create the hop line") CRLF);
      D(perror("rfc_received"));
      exit_return = CMD_RETURN_FAILURE;
      goto exit_failure;
   }
   state->data_buf_pnt = state->data_buf;
   state->data_buf_size = DATA_CHUNK_SIZE;
   state->data_size = (size_t) 0U;
   state->eof_status = DATA_INITIAL_EOF_STATUS;
   state->header_status = DATA_INITIAL_HEADER_STATUS;
   state->header_received_pnt = HEADER_RECEIVED;
   state->header_delivered_pnt = HEADER_DELIVERED;      
   state->hops = 0U;
   if ((received_pnt = received->str) != NULL) {
      while (*received_pnt != 0) {
	 if (++(state->data_size) > config.databytes) {
	    D(puts("size exceeded"));
	    say_flush(state, _("552 Message too long") CRLF);
	    exit_return = CMD_RETURN_FAILURE;
	    goto exit_failure;
	 }	 
	 if (data_bufferize(*received_pnt, state) < 0) {
	    D(puts("data_bufferize"));
	    say_flush(state, _("421 Write error") CRLF);
	    exit_return = CMD_RETURN_FAILURE;
	    goto exit_failure;	    
	 }
	 received_pnt++;
      }
   }

   goto exit_normal;
   
   exit_failure :
   if (state->data_buf != NULL) {
      free(state->data_buf);
      state->data_buf = 0;
   }
   exit_normal :
   if (received != NULL) {
      var_string_free(received);
   }
   return exit_return;
}

static const CmdList commands[] = {
   { "QUIT", cmd_quit },
   { "HELO", cmd_helo },
   { "EHLO", cmd_ehlo },
   { "NOOP", cmd_noop },
   { "VRFY", cmd_vrfy },
   { "EXPN", cmd_expn },
   { "RSET", cmd_rset },
   { "MAIL", cmd_mail },
   { "RCPT", cmd_rcpt },
   { "SIZE", cmd_size },   
   { "DATA", cmd_data },
};

static int say_unknown(State * const state)
{
   say_flush(state, _("502 Unimplemented") CRLF);
   
   return CMD_RETURN_UNKNOWN;
}

static int parse_cmd(char *cmd, State * const state)
{
   char *arg = NULL;
   const CmdList *commands_pnt = commands;
   size_t c = (sizeof commands / sizeof commands[0]);
   
   cmd = skip_spaces(cmd);
   if ((arg = strchr(cmd, ' ')) != NULL) {
      if (*arg != 0) {
	 *arg = 0;
	 arg++;
	 arg = skip_spaces(arg);
      } 
      if (*arg == 0) {
	 arg = NULL;
      }
   }
   do {
      if (strcasecmp(commands_pnt->cmd, cmd) == 0) {
	 return commands_pnt->func(arg, state);
      }
      commands_pnt++;
      c--;
   } while (c != 0U);
   say_unknown(state);
   
   return CMD_RETURN_UNKNOWN;
}

static int say_helo(State * const state)
{
   say(state, "220 ");
   say(state, config.smtpgreeting);
   say_flush(state, " ESMTP" CRLF);
   
   return 0;
}

#ifdef LAZY_METADATA_SYNC
void *sync_thread(void *thread_args_)
{
   SyncNode sync_node;
   int flag;

   (void) thread_args_;
   D(puts("sync_thread started"));
   for(;;) {
      D(puts("sync_thread is looking for something to commit"));
      sleep(LAZY_SYNC_DELAY);
      flag = 0;     
      while (sync_queue_pop(&global_sync_queue, &sync_node) == 0) {
	 some_more:
# ifdef INTD_COMPATIBILITY
	 D(printf("sync_thread popped fd [%d] and fd [%d]\n" __
		  sync_node.fd1, sync_node.fd2));
# else
	 D(printf("sync_thread popped fd [%d]\n" __
		  sync_node.fd1));	 
# endif
	 if (sync_node.fd1 >= 0) {
	    fsync(sync_node.fd1);
	    close(sync_node.fd1);
	 }
# ifdef INTD_COMPATIBILITY
	 if (sync_node.fd2 >= 0) {
	    fsync(sync_node.fd2);
	    close(sync_node.fd2);	    
	 }
# endif
	 flag = 1;
      }      
      if (sync_queue_pop(&global_sync_queue, &sync_node) == 0) {
	 D(puts("sync_thread some_more"));
	 goto some_more;
      }
      semctl(sync_semid, 0, SETVAL, 0);
      if (flag != 0) {
# ifdef INTD_COMPATIBILITY
	 trigger_qmail_send();
# else
	 trigger_lmail_send();
# endif
      }
   }
   
   return NULL;
}
#endif

void *client_thread(void *thread_args_)
{
   ThreadArgs *thread_args = (ThreadArgs *) thread_args_;
   BufferedReader client_br;
   BufferedWriter client_bw;   
   char line[LINE_MAX];
   State state;
   int cmd_ret;
   char c;

   D(puts("New running client thread"));
   if (thread_args == NULL) {
      D(puts("No arg to the client thread ?!"));
      return NULL;
   }
   if (thread_args->client_fd < 0) {
      D(puts("No valid file descriptor passed to the thread"));
      return NULL;
   }
   init_state(&state, 0);
   buffered_reader_init(&client_br, thread_args->client_fd);
   buffered_writer_init(&client_bw, thread_args->client_fd);   
   free(thread_args);   			
   state.client_br = &client_br;
   state.client_bw = &client_bw;   
   
   say_helo(&state);
   for (;;) {
      for (;;) {
	 if (buffered_reader_get_line(&client_br, line, sizeof line) <= 0) {
	    goto bye;
	 }
	 chomp(line);
	 trim_rspaces(line);
	 cmd_ret = parse_cmd(line, &state);
	 if (cmd_ret == CMD_RETURN_QUIT) {
	    goto bye;
	 } else if (cmd_ret == CMD_RETURN_INVALID) {
	    say_flush(&state, _("-- Please stop playing with telnet on port 25 --") CRLF);
	    goto bye;
	 } else if (cmd_ret == CMD_RETURN_FAILURE) {
	    D(perror("Failure"));
	    goto bye;
	 }
	 if (state.data_buf != NULL) {
	    break;
	 }
      }
      D(puts("Entering in data state"));      
      while (state.data_buf != NULL) {
	 if (buffered_reader_readchar(&client_br, &c) < 0) {
	    goto bye;
	 }
	 if ((cmd_ret = cmd_data_fetch(c, &state)) != 0) {
	    break;
	 }
      }
      free_state(&state, 1);
      init_state(&state, 1);
   }
   bye:
   buffered_writer_close(&client_bw, 0);   
   buffered_reader_close(&client_br, 1);
   free_state(&state, 0);
   total_clients--;
   
   return NULL;
}

int main(void)
{
   int kindy;
   struct sockaddr_in sa_server;
#ifdef LAZY_METADATA_SYNC
   st_thread_t sync_thread_id;
#endif

#ifndef NO_STATE_THREAD
      {
	 pid_t pid;
	 
	 if ((pid = fork()) < 0) {
	    perror(_("fork"));
	    exit(EXIT_FAILURE);
	 }
	 if (pid > 0) {
	    exit(EXIT_SUCCESS);
	 }
	 setsid();
	 if ((pid = fork()) < 0) {
	    perror(_("fork"));
	    exit(EXIT_FAILURE);
	 }
	 if (pid > 0) {
	    exit(EXIT_SUCCESS);
	 }	 
      }
#endif
   
   if (st_init() < 0) {
      perror(_("Unable to use state threads"));
      return -1;
   }
         
   setvbuf(stdin, NULL, _IONBF, 1U);
   setvbuf(stdout, NULL, _IONBF, 1U);
   umask(003);      
   atexit(ipc_delete);
   signal(SIGHUP, sig);
   signal(SIGINT, sig);   
   signal(SIGQUIT, sig_term);
#ifdef SIGILL   
   signal(SIGILL, sig);
#endif
   signal(SIGFPE, sig);            
   signal(SIGKILL, sig_term);               
   signal(SIGPIPE, sig);
   signal(SIGTERM, sig_term);
   signal(SIGUSR1, sig);      
   signal(SIGUSR2, sig);

   check_fqdn_consistency_init();
   if (read_config_files() < 0) {
      perror(_("Error when reading configuration files"));
      return -1;
   }
   if ((kindy = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
      perror(_("socket"));
      return -1;
   }
     {
	int opt = 1;
	
	if (setsockopt(kindy, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof opt) < 0) {
	   D(perror("setsockopt"));
	}
     }
   memset(&sa_server, 0, sizeof sa_server);
   sa_server.sin_family = AF_INET;
   sa_server.sin_port = htons(config.smtpdport);
   sa_server.sin_addr.s_addr = INADDR_ANY;
   if (bind(kindy, (struct sockaddr *) &sa_server, sizeof sa_server) < 0) {
      perror("bind");
      return -1;
   }   
   if (change_id(LMAIL_SMTPD_USER, LMAIL_SMTPD_GROUP) < 0) {
      puts(_("Sorry, but I need to change my id to ")
	     LMAIL_SMTPD_USER "/" LMAIL_SMTPD_GROUP " (uid/gid).\n"
	   _("Please run me as root and/or check that this virtual user exists."));
      return -1;
   }
   if (init_sequence() < 0) {
      perror("init_sequence");
      return -1;
   }
#ifdef LAZY_METADATA_SYNC
   if (sync_queue_init(&global_sync_queue) < 0) {
      perror("sync_queue_init");
      return -1;
   }   
   if ((sync_semid = semget(IPC_PRIVATE, 1, 0600)) < 0) {
      D(perror("semget"));
      return -1;
   }   
   if ((sync_thread_id = st_thread_create(sync_thread, NULL, 0, 0)) = NULL) {
      perror("st_thread_create");
      return -1;
   }
#endif
   if (listen(kindy, BACKLOG) < 0) {
      perror("listen");
      return -1;
   }
   for (;;) {
      int client_fd;
      struct sockaddr_in sa_client;
      socklen_t sa_client_size;
      st_thread_t thread;
#ifdef PREALLOCATE
      st_thread_t preallocate_thread;
#endif
      ThreadArgs *thread_args;

      if (server_shutdown != 0) {
	 close(kindy);
	 for (;;) {
	    pause();
	    if (server_shutdown == 2) {
	       break;
	    }
	 }
      }
#ifdef PREALLOCATE
      D(printf("before preallocation, total_clients = [%d]\n" __
	       total_clients));
      if (total_clients < 2) {
	 if ((preallocate_thread = 
	      st_thread_create(preallocate, NULL, 1, 0)) == NULL) {
	    D(puts("st_thread_create"));	    
	 }
      }
#endif
      D(puts("before accept"));
      client_fd = accept(kindy, (struct sockaddr *) &sa_client, &sa_client_size);
#ifdef PREALLOCATE
      total_clients++;
      st_thread_join(preallocate_thread, NULL);
      total_clients--;
#endif
      if (client_fd < 0) {
	 D(perror("accept"));
	 continue;
      }
      setsockopt(client_fd, IPPROTO_IP, IP_OPTIONS, NULL, (socklen_t) 0);
      if ((thread_args = malloc(sizeof *thread_args)) == NULL) {
	 D(perror("malloc"));
	 close(client_fd);
	 continue;
      }
      thread_args->client_fd = client_fd;
      D(puts("New connection accepted"));
      if (total_clients < config.max_clients) {
	 total_clients++;
	 if ((thread = st_thread_create(client_thread, thread_args, 0, 0)) == NULL) {
	    D(puts("st_thread_create failure"));
	    free(thread_args);
	    close(client_fd);
	    continue;
	 }
	D(puts("new thread created"));
      } else {
	 D(puts("Too many connections"));
	 total_clients++;
	 client_thread(thread_args);	 
      }
      D(puts("Thread should be running"));
      if (total_clients > config.max_clients) {
	 total_clients = config.max_clients;
      } else if (total_clients < 0) {
	 total_clients = 0;
      }
   }
   if (config.smtpgreeting != NULL) {
      free((void *) config.smtpgreeting);
      config.smtpgreeting = NULL;
   }
   if (config.has_badmailfrom != 0) {
      hashed_list_destroy(&config.badmailfrom);
      config.has_badmailfrom = 0;
   }
   if (config.open_relay == 0) {
      hashed_list_destroy(&config.rcpthosts_strict);
      hashed_list_destroy(&config.rcpthosts_joker);
      config.open_relay = 1;
   }
   if (config.morercpthosts.fd >= 0) {
      configfile_free_cdb(&config.morercpthosts);
      config.morercpthosts.fd = -1;
   }
   st_thread_exit(NULL);
   
   return 0;
}
